# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Run an annual irradiance study for a Honeybee model to compute hourly solar
irradiance for each sensor in a model's sensor grids.
_
The fundamental calculation of this recipe is the same as that of "HB Annual
Daylight" in that an enhaced 2-phase method is used to accurately account for
direct sun at each simulation step. However, this recipe computes broadband
solar irradiance in W/m2 instead of visible illuminance in lux.
_
Consequently, the average irradiance and cumulative radiation values produced from
this recipe are more accurate than those produced by the "HB Cumulative Radiation"
recipe. Furthermore, because the hourly irriadiance values are accurate, this
recipe can be used to evaluate `peak_irradiance` and determine the worst-case
solar loads over clear sky Weas that represent cooling design days.

-
    Args:
        _model: A Honeybee Model for which Annual Irradiance will be simulated.
            Note that this model must have grids assigned to it.
        _wea: A Wea object produced from the Wea components that are under the Light
            Sources tab. This can also be the path to a .wea or a .epw file.
        _timestep_: An integer for the timestep of the inpput _wea. This value is used
            to compute average irradiance and cumulative radiation. (Default: 1)
        visible_: Boolean to indicate the type of irradiance output, which can be solar (False)
            or visible (True). Note that the output values will still be
            irradiance (W/m2) when "visible" is selected but these irradiance
            values will be just for the visible portion of the electromagnetic
            spectrum. The visible irradiance values can be converted into
            illuminance by multiplying them by the Radiance luminous efficacy
            factor of 179. (Default: False).
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        grid_filter_: Text for a grid identifer or a pattern to filter the sensor grids of
            the model that are simulated. For instance, first_floor_* will simulate
            only the sensor grids that have an identifier that starts with
            first_floor_. By default, all grids in the model will be simulated.
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        results: Raw result files (.ill) that contain matrices of irradiance in W/m2
            for each time step of the wea.
        res_direct: Raw result files (.ill) that contain irradiance matrices for just the
            direct sun at each hour of the simulation. These can be postprocessed
            using various components under the 4::Results sub-tab.
        avg_irr: The average irradiance in W/m2 for each sensor over the Wea time period.
        peak_irr: The highest irradiance value in W/m2 during the Wea time period. This
            is suitable for assessing the worst-case solar load of clear skies on
            cooling design days. It can also be used to determine the highest
            radiant temperatures that occupants might experience in over the
            time period of the Wea.
        radiation: The cumulative radiation in kWh/m2 over the Wea time period.
"""

ghenv.Component.Name = 'HB Annual Irradiance'
ghenv.Component.NickName = 'AnnualIrradiance'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '3 :: Recipes'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from lbt_recipes.recipe import Recipe
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('annual-irradiance')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('wea', _wea)
    recipe.input_value_by_name('timestep', _timestep_)
    recipe.input_value_by_name('output-type', visible_)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('grid-filter', grid_filter_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(run_settings_, radiance_check=True, silent=silent)

    # load the results
    try:
        results = recipe_result(recipe.output_value_by_name('results', project_folder))
        res_direct = recipe_result(recipe.output_value_by_name('results-direct', project_folder))
        avg_irr = recipe_result(recipe.output_value_by_name('average-irradiance', project_folder))
        peak_irr = recipe_result(recipe.output_value_by_name('peak-irradiance', project_folder))
        radiation = recipe_result(recipe.output_value_by_name('cumulative-radiation', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct a Wea object into data collections of direct, diffuse, and golbal
irradiance at each timestep of the file.

-
    Args:
        _wea: A Honeybee WEA object.

    Returns:
        readMe!: Reports, errors, warnings, etc.
        dir: A data collection of direct normal irradiance values at each timestep
            of the Wea.
        diff: A data collection of diffuse sky solar irradiance values at each
            timestep of the Wea.
        glob: A data collection of global horizontal irradiance values at each
            timestep of the Wea.
"""

ghenv.Component.Name = 'HB Deconstruct Wea'
ghenv.Component.NickName = 'DecnstrWea'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '2 :: Light Sources'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    dir = _wea.direct_normal_irradiance
    diff = _wea.diffuse_horizontal_irradiance
    glob = _wea.global_horizontal_irradiance


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a list of exterior modifiers that can be used to edit or create a
ModifierSet object.
-

    Args:
        _exterior_wall_: A modifier object for exterior walls (or text for
            the identifier of the modifier within the library).
        _exterior_roof_: A modifier object for exterior roofs (or text for
            the identifier of the modifier within the library).
        _exposed_floor_: A modifier object for exposed floors (or text for
            the identifier of the modifier within the library).
    
    Returns:
        exterior_set: A list of exterior modifiers that can be used to edit
            or create a ModifierSet object.
"""

ghenv.Component.Name = 'HB Exterior Modifier Subset'
ghenv.Component.NickName = 'ExteriorSubset'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '1 :: Modifiers'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import honeybee_radiance dependencies
    from honeybee_radiance.modifier import Modifier
    from honeybee_radiance.lib.modifiers import modifier_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))


def check_mod(mod, input_name):
    """Get an Modifier from the library if it's a string."""
    if isinstance(mod, str):
        return modifier_by_identifier(mod)
    else:
        assert isinstance(mod, Modifier), \
            'Expected Modifier for {}. Got {}'.format(input_name, type(mod))
    return mod


# go through each input modifier
if _exterior_wall_ is not None:
    _exterior_wall_ = check_mod(_exterior_wall_, '_exterior_wall_')
if _exterior_roof_ is not None:
    _exterior_roof_ = check_mod(_exterior_roof_, '_exterior_roof_')
if _exposed_floor_ is not None:
    _exposed_floor_ = check_mod(_exposed_floor_, '_exposed_floor_')


# return the final list from the component
exterior_set = [_exterior_wall_, _exterior_roof_, _exposed_floor_]



# Honeybee: A Plugin for Environmental Analysis
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Get the recommended ambient resoluation (-ar) needed to resolve details with a
given dimension in model units.
_
This recommendation is derived from the overall dimensions of the Radince scene
being simulated as well as the ambient accuracy (-aa) being used in the simulation.
_
The result from this component can be plugged directly into the additional_par_
of the "HB Radiance Parameter" component or into the radiance_par of any
recipe components.

-
    Args:
        _model: The Honeybee Model being used for Radiance simulation.
        _detail_dim: A number in model units that represents the dimension of the
            smallest detail that must be resolved in the Radiance simulation.
        _aa_: An number for ambient accuracy (-aa) being used in the Radiance smiulation.
            This value should be matched between this component and the component
            into which the ouput ar is being input. (Default: 0.25 for
            low-resolution Radiance studies).

    Returns:
        ar: The abmient resolution needed to resolve the _detail_dim as a text string.
            These can be plugged into the additional_par_ of the "HB Radiance
            Parameter" component or the radiance_par_ input of the recipes.
"""

ghenv.Component.Name = 'HB Ambient Resolution'
ghenv.Component.NickName = 'AR'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '3 :: Recipes'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default -aa
    aa = 0.25 if _aa_ is None else _aa_

    # get the longest dimension
    min_pt, max_pt = _model.min, _model.max
    x_dim = max_pt.x - min_pt.x
    y_dim = max_pt.y - min_pt.y
    z_dim = max_pt.z - min_pt.z
    longest_dim = max((x_dim, y_dim, z_dim))
    
    # calculate the ambient resolution.
    a_res = int((longest_dim * aa) / _detail_dim)
    ar = '-ar {}'.format(a_res)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Run an annual glare study for a Honeybee model to compute hourly Daylight Glare
Probability (DGP) for each sensor in a model's sensor grids.
_
This recipe uses the image-less glare method developed by Nathaniel Jones to
estimate glare at each sensor. More information on this method can be found here:
https://github.com/nljones/Accelerad/wiki/The-Imageless-Method-for-Spatial-and-Annual-Glare-Analysis
_
The resulting DGP is used to compute Glare Autonomy (GA), which is the percentage
of occupied time that a view is free of glare.

-
    Args:
        _model: A Honeybee Model for which Annual Daylight Glare Probability (DGP) will
            be simulated. Note that this model must have grids assigned to
            it and, typically, these are radial grids created using the
            "radial grid" components.
        _wea: A Wea object produced from the Wea components that are under the Light
            Sources tab. This can also be the path to a .wea or a .epw file.
            Note that the Wea must have a timestep of 1 to be used with this
            recipe.
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        _glare_thresh_: A fractional number for the threshold of DGP above which conditions
            are considered to induce glare. This value is used when calculating
            glare autonomy, which is the percent of hours in which the view is
            free of glare. (Default: 0.4 for disturbing or intolerable glare).
        _luminance_fac_: Luminance factor in cd/m2. If the sky patch brightness
            is above this factor it will act as a glare source. (Default: 2000).
        _schedule_: An annual occupancy schedule, either as a Ladybug Hourly Continuous
            Data Collection or a HB-Energy schedule object. This can also be the
            path to a CSV file with 8760 rows or the identifier of a schedule in
            the honeybee-energy schedule library. Any value in this schedule
            that is 0.1 or above will be considered occupied.
        grid_filter_: Text for a grid identifer or a pattern to filter the sensor grids of
            the model that are simulated. For instance, first_floor_* will simulate
            only the sensor grids that have an identifier that starts with
            first_floor_. By default, all grids in the model will be simulated.
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        results: Raw result files (.dgp) that contain Daylight Glare Probability (DGP)
            matrices for each sensor at each hour of the simulation. These can
            be postprocessed using various components under the 4::Results sub-tab.
        GA: Glare Autonomy (GA) results in percent. GA is the percentage of occupied hours
            that each view is free of glare (with a DGP below the glare threshold).
            These can be plugged into the "LB Spatial Heatmap" component along
            with meshes of the sensor grids to visualize results.
"""

ghenv.Component.Name = 'HB Imageless Annual Glare'
ghenv.Component.NickName = 'AnnualGlare'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '3 :: Recipes'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from lbt_recipes.recipe import Recipe
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('imageless-annual-glare')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('wea', _wea)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('glare-threshold', _glare_thresh_)
    recipe.input_value_by_name('luminance-factor', _luminance_fac_)
    recipe.input_value_by_name('schedule', _schedule_)
    recipe.input_value_by_name('grid-filter', grid_filter_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(run_settings_, radiance_check=True, silent=silent)

    # load the results
    try:
        results = recipe_result(recipe.output_value_by_name('results', project_folder))
        GA = recipe_result(recipe.output_value_by_name('ga', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2022, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate Annual Sunlight Exposure from a results folder.
_
Note: This component will only output a LEED compliant ASE if you've run the
simulation with all operable shading devices retracted. If you are using
results with operable shading devices, then this output is NOT LEED compliant.

-
    Args:
        _results: An annual results folder containing direct illuminance results.
            This can be the output of the "HB Annual Daylight" component. This can
            also be just the path to the results folder.
        _occ_sch_: An annual occupancy schedule as a Ladybug Data Collection or a HB-Energy
            schedule object. This can also be the identifier of a schedule in
            your HB-Energy schedule library. Any value in this schedule that is
            0.1 or above will be considered occupied. If None, a schedule from
            9AM to 5PM on weekdays will be used.
        _threshold_: The threshold (lux) that determines if a sensor is
            overlit (default: 1000).
        _target_hrs_: The number of occupied hours that cannot receive higher
            illuminance than the direct threshold (default: 250).
        grid_filter_: The name of a grid or a pattern to filter the grids. For instance,
            first_floor_* will simulate only the sensor grids that have an
            identifier that starts with first_floor_. By default all the grids
            will be processed.

    Returns:
        report: Reports, errors, warnings, etc.
        ASE: Annual sunlight exposure as a percentage for each sensor grid.
        hrs_above_thresh: The number of hours above the threshold for each sensor point.
            These can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize results.
"""

ghenv.Component.Name = "HB Annual Sunlight Exposure"
ghenv.Component.NickName = 'ASE'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os
import subprocess

try:
    from ladybug.datacollection import BaseCollection
    from ladybug.futil import write_to_file
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:  # honeybee schedule library is not available
    schedule_by_identifier = None

try:
    from pollination_handlers.outputs.daylight import read_ase_from_folder, \
        read_hours_from_folder
except ImportError as e:
    raise ImportError('\nFailed to import pollination_handlers:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set default values for the thresholds and the grid filter
    grid_filter_ = '*' if grid_filter_ is None else grid_filter_
    _direct_threshold_ = _threshold_ if _threshold_ else 1000
    _occ_hours_ = _target_hrs_ if _target_hrs_ else 250

    # process the schedule
    if _occ_sch_ is None:
        schedule = None
    elif isinstance(_occ_sch_, BaseCollection):
        schedule = _occ_sch_.values
    elif isinstance(_occ_sch_, str):
        if schedule_by_identifier is not None:
            try:
                schedule = schedule_by_identifier(_occ_sch_).values()
            except TypeError:  # it's probably a ScheduleFixedInterval
                schedule = schedule_by_identifier(_occ_sch_).values
        else:
            raise ValueError('honeybee-energy must be installed to reference '
                             'occupancy schedules by identifier.')
    else:  # assume that it is a honeybee schedule object
        try:
            schedule = _occ_sch_.values()
        except TypeError:  # it's probably a ScheduleFixedInterval
            schedule = _occ_sch_.values
    if schedule is not None:
        bin_schedule = []
        for val in schedule:
            bin_val = 1 if val >= 0.1 else 0
            bin_schedule.append(bin_val)
        schedule = bin_schedule

    # compute the annual metrics
    res_folder = os.path.dirname(_results[0]) if os.path.isfile(_results[0]) \
        else _results[0]
    if os.path.isfile(os.path.join(res_folder, 'grid_states.json')):
        cmds = [
            folders.python_exe_path, '-m', 'honeybee_radiance_postprocess',
            'post-process', 'annual-sunlight-exposure', res_folder, '-sf', 'metrics',
            '-dt', str(_direct_threshold_), '-oh', str(_occ_hours_)
        ]
        if grid_filter_ != '*':
            cmds.extend(['--grids-filter', grid_filter_])
        if schedule is not None:
            sch_str = '\n'.join(str(h) for h in schedule)
            sch_file = os.path.join(res_folder, 'schedule.txt')
            write_to_file(sch_file, sch_str)
            cmds.extend(['--schedule', sch_file])
        use_shell = True if os.name == 'nt' else False
        process = subprocess.Popen(
            cmds, cwd=res_folder, shell=use_shell,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout = process.communicate()  # wait for the process to finish
        if stdout[-1] != '':
            print(stdout[-1])
            raise ValueError('Failed to compute annual sunlight exposure.')
        metric_dir = os.path.join(res_folder, 'metrics')
        ASE = read_ase_from_folder(os.path.join(metric_dir, 'ase'))
        hrs_above_thresh = list_to_data_tree(read_hours_from_folder(os.path.join(metric_dir, 'hours_above')))
    else:
        raise ValueError('Invalid results folder!')



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Write a Honeybee Model to a Radiance Model Folder.
_
This Radiance Model Folder is what is used to run various types of Radiance
studies off of a consistent set of geometry and modifiers.

-

    Args:
        _model: A honeybee model object possessing all geometry, radiance modifiers
            and simulation assets like Sensor Grids and Views.
        minimal_: Boolean to note whether the radiance strings should be written in a minimal
            format (with spaces instead of line breaks). (Default: False).
        _folder_: Path to a folder to into which the Model Radiance Folder will be
            written. If unspecified, it will be written to a sub-folder
            within the default simulation folder.
        _write: Set to True to write the Model to a Radiance folder.

    Returns:
        report: Reports, errors, warnings, etc.
        folder: Path to the folder in which all of the files have been written.
"""

ghenv.Component.Name = 'HB Model to Rad Folder'
ghenv.Component.NickName = 'ModelToRad'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import os
import re

try:
    from ladybug.futil import write_to_file_by_name, nukedir, preparedir
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_radiance_folder.folder import ModelFolder
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance_folder:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _write:
    # process the simulation folder name and the directory
    clean_name = re.sub(r'[^.A-Za-z0-9_-]', '_', _model.display_name)
    folder = os.path.join(folders.default_simulation_folder, clean_name, 'radiance') \
        if _folder_ is None else _folder_
    if os.path.isdir(folder):
        nukedir(folder, rmdir=True)  # delete the folder if it already exists
    else:
        preparedir(folder)  # create the directory if it's not there

    # write the model folder
    _model.to.rad_folder(_model, folder, minimal=bool(minimal_))



# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize a sky as a High Dynamic Range (HDR) image file.
-

    Args:
        _sky: A Radiance sky from any of the sky components under the "Light Sources" tab.
            Text string representations of skies are also acceptable.
        _size_: A number for the X and Y dimension of the imgae in pixles. (Default: 500 px)

    Returns:
        hdr: Path to the High Dynamic Range (HDR) image file of the sky. This can be
            plugged into the Ladybug "Image Viewer" component to preview the image.
            It can also be plugged into the "HB False Color" component to convert
            the image into a false color version. Lastly, it can be connected to
            the "HB HDR to GIF" component to get a GIF image that is more portable
            and easily previewed by different software.
        ghi: The global horizontal irradiance (W/m2) for an upstructed test point under
            the sky.
"""

ghenv.Component.Name = 'HB Visualize Sky'
ghenv.Component.NickName = 'VizSky'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '2 :: Light Sources'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import os

try:  # import honeybee_radiance dependencies
    from ladybug.futil import write_to_file_by_name
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import honeybee dependencies
    from honeybee.config import folders
    from honeybee.typing import clean_rad_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import honeybee_radiance_command dependencies
    from honeybee_radiance_command.oconv import Oconv
    from honeybee_radiance_command.rpict import Rpict
    from honeybee_radiance_command.rtrace import Rtrace
    from honeybee_radiance_command.rcalc import Rcalc
    from honeybee_radiance_command.pcond import Pcond
    from honeybee_radiance_command.pflip import Pflip
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance_command:\n\t{}'.format(e))

try:  # import honeybee_radiance dependencies
    from honeybee_radiance.config import folders as rad_folders
    from honeybee_radiance.lightsource.sky.strutil import string_to_sky
    from honeybee_radiance.lightsource.sky import ClimateBased
    from honeybee_radiance.sensorgrid import SensorGrid
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# check the Radiance date of the installed radiance
try:  # import lbt_recipes dependencies
    from lbt_recipes.version import check_radiance_date
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))
check_radiance_date()


if all_required_inputs(ghenv.Component):
    # set defaults and process the sky input
    _size_ = 500 if _size_ is None else _size_
    if isinstance(_sky, str):  # convert the sky string into a sky object
        _sky = string_to_sky(_sky)
    sky_content = _sky.to_radiance(1) if isinstance(_sky, ClimateBased) else _sky.to_radiance()

    # set up the paths for the various files used in translation
    sky_dir = os.path.join(folders.default_simulation_folder, 'sky_visualiztion')
    sky_file, sky_oct = 'weather.sky', 'sky_visual.oct'
    write_to_file_by_name(sky_dir, sky_file, sky_content, mkdir=True)
    ghi_res, full_ghi_res = 'ghi.res', os.path.join(sky_dir, 'ghi.res')
    init_hdr, final_hdr = 'sky_init.HDR', '{}.HDR'.format(clean_rad_string(str(_sky)))
    hdr = os.path.join(sky_dir, final_hdr)
    if os.path.isfile(hdr):
        os.remove(hdr)

    # build up the commands to render the image of the sky
    oconv = Oconv(inputs=[sky_file], output=sky_oct)
    oconv.options.f = True

    rpict = Rpict(octree=sky_oct, output=init_hdr)
    rpict.options.i = True
    rpict.options.t = 10
    rpict.options.ab = 1
    rpict.options.ad = 1000
    rpict.options.as_ = 20
    rpict.options.ar = 300
    rpict.options.aa = 0.1
    rpict.options.x = _size_
    rpict.options.y = _size_
    rpict.options.vt = 'h'
    rpict.options.vp = (0, 0, 0)
    rpict.options.vd = (0, 0, 1)
    rpict.options.vu = (0, 1, 0)
    rpict.options.vh = 180
    rpict.options.vv = 180

    pflip = Pflip(input=init_hdr, output=final_hdr)
    pflip.options.h = True

    # add the command to get the horizontal irradiance of the sky
    grid = SensorGrid.from_position_and_direction('up_sensor', [(0, 0, 0)], [(0, 0, 1)])
    grid.to_file(sky_dir, 'up_sensor.pts')
    rtrace = Rtrace(octree=sky_oct, sensors='up_sensor.pts')
    rtrace.options.I = True
    rtrace.options.w = True
    rtrace.options.h = True
    rtrace.options.ab = 1
    rcalc = Rcalc(output=ghi_res)
    rcalc.options.e = '$1=(0.265*$1+0.67*$2+0.065*$3)'
    rtrace.pipe_to = rcalc

    # run the commands in series and load the global horizontal irradiance
    env = None
    if rad_folders.env != {}:
        env = rad_folders.env
    env = dict(os.environ, **env) if env else None
    for r_cmd in (oconv, rpict, pflip, rtrace):
        r_cmd.run(env, cwd=sky_dir)
    with open(full_ghi_res, 'r') as inf:
        ghi = inf.readlines()[0].strip()



# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Import the hourly illuminance or irradiance results of an annual daylight or irradiance
study to a list of data collections.
_
The resulting data collections can be visulized using the ladybug components or
deconstructed for detailed analysis with native Grasshopper math components.

-
    Args:
        _results: An list of annual Radiance result files from the "HB Annual Daylight"
            component (containing the .ill files and the sun-up-hours.txt).
            This can also be just the path to the folder containing these
            result files.
        _sel_pts: A point or list of points, which will be used to filter the sensors
            for which data collections will be imported.
        _all_pts: The data tree of all sensor points that were used in the simulation.
            This is required in order to look up the index of the _sel_pts in
            the results matrices.
        sel_vecs_: An optional vector or list of vectors, which will be used to filter
            the sensors for which data collections will be imported. If there
            is an input here, the all_vecs_ must be connected.
        all_vecs_: The data tree of all sensor directions that were used in the simulation.
            This is required in order to look up the index of the sel_vecs_ in
            the results matrices. 

 Returns:
        report: Reports, errors, warnings, etc.
        data: A list of hourly data collections containing illuminance or irradiance results.
            These can be visulized using the ladybug components or deconstructed
            for detailed analysis with native Grasshopper math components.
"""

ghenv.Component.Name = 'HB Annual Results to Data'
ghenv.Component.NickName = 'AnnualToData'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os
import json
import subprocess

try:
    from ladybug.datatype.illuminance import Illuminance
    from ladybug.datatype.energyflux import Irradiance
    from ladybug.datatype.time import Time
    from ladybug.datatype.fraction import Fraction
    from ladybug.analysisperiod import AnalysisPeriod
    from ladybug.header import Header
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.futil import write_to_file
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_radiance.postprocess.annualdaylight import _process_input_folder
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import tolerance
    from ladybug_rhino.togeometry import to_point3d, to_vector3d
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        data_tree_to_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def file_to_data(ill_file, point_filter, su_pattern, header, timestep, grid_id):
    """Get a list of data collections for a given result file."""
    data_colls = []
    new_header = header.duplicate()
    new_header.metadata['sensor grid'] = grid_id
    with open(ill_file) as results:
        if point_filter is None:
            for pt_res in results:
                base_values = [0] * 8760 * timestep
                for val, hr in zip(pt_res.split(), su_pattern):
                    base_values[hr] = float(val)
                data_colls.append(HourlyContinuousCollection(new_header, base_values))
        else:
            for i, pt_res in enumerate(results):
                if i in point_filter:
                    new_header = new_header.duplicate()
                    new_header.metadata['sensor index'] = i
                    base_values = [0] * 8760 * timestep
                    for val, hr in zip(pt_res.split(), su_pattern):
                        base_values[hr] = float(val)
                    data_colls.append(HourlyContinuousCollection(new_header, base_values))
    return data_colls


def find_point_in_grid(s_pt, all_pts):
    """Find the index of a point in a list of list of grids."""
    m_pts = []
    for i, grid_pts in enumerate(all_pts):
        for j, g_pt in enumerate(grid_pts):
            if g_pt.is_equivalent(s_pt, tolerance):
                m_pts.append((i, j))
    return m_pts


def find_vec_in_grid(s_v, all_vecs, pt_filter):
    """Find the index of a vector in a list of list of grids."""
    m_vecs = []
    for i, grid in enumerate(pt_filter):
        for j in grid:
            if all_vecs[i][j].is_equivalent(s_v, tolerance):
                m_vecs.append((i, j))
    return m_vecs


if all_required_inputs(ghenv.Component):
    # get the relevant .ill files
    res_folder = os.path.dirname(_results[0]) if os.path.isfile(_results[0]) \
        else _results[0]
    grids, sun_up_hours = _process_input_folder(res_folder, '*')

    # set up the sensor filter
    pt_filter = [None for i in grids]
    if len(_sel_pts) != 0 or len(sel_vecs_) != 0:
        pt_filter = [[] for i in grids]

    # check the sel_pts and all_pts input
    if len(_sel_pts) != 0:
        all_pts = [[to_point3d(pt) for pt in dat[-1]] for dat in data_tree_to_list(_all_pts)]
        assert len(all_pts) != 0, '_all_pts must be connected in order to use _sel_pts.'
        sel_pts = [to_point3d(pt) for pt in _sel_pts]
        for s_pt in sel_pts:
            m_pts = find_point_in_grid(s_pt, all_pts)
            for i, j in m_pts:
                pt_filter[i].append(j)

    # check the sel_vecs and all_vecs input
    if len(sel_vecs_) != 0:
        new_pt_filter = [[] for i in grids]
        all_vecs = [[to_vector3d(v) for v in dat[-1]] for dat in data_tree_to_list(all_vecs_)]
        assert len(all_vecs) != 0, 'all_vecs_ must be connected in order to use sel_vecs_.'
        sel_vecs = [to_vector3d(v) for v in sel_vecs_]
        for s_v in sel_vecs:
            m_vs = find_point_in_grid(s_v, all_vecs) if len(_sel_pts) == 0 else \
                find_vec_in_grid(s_v, all_vecs, pt_filter)
            for i, j in m_vs:
                new_pt_filter[i].append(j)
        pt_filter = new_pt_filter

    # check to see if results use the newer numpy arrays
    if os.path.isdir(os.path.join(res_folder, '__static_apertures__')):
        cmds = [folders.python_exe_path, '-m', 'honeybee_radiance_postprocess',
                'post-process', 'annual-to-data', res_folder]
        if pt_filter[0] is not None:
            sen_dict = {g['full_id']: s_ind for g, s_ind in zip(grids, pt_filter)}
            si_file = os.path.join(res_folder, 'sensor_indices.json')
            write_to_file(si_file, json.dumps(sen_dict))
            cmds.extend(['--sensor-index', si_file])
        use_shell = True if os.name == 'nt' else False
        process = subprocess.Popen(
            cmds, cwd=res_folder, shell=use_shell,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()  # wait for the process to finish
        if stderr != '':
            print(stderr)
            raise ValueError('Failed to compute data collections.')
        data_dicts = json.loads(stdout)
        data = [[HourlyContinuousCollection.from_dict(d) for d in data]
                for data in data_dicts]
        data = list_to_data_tree(data)

    else:
        # extract the timestep if it exists
        timestep, has_t_step = 1, False
        tstep_file = os.path.join(res_folder, 'timestep.txt')
        if os.path.isfile(tstep_file):  # it's an annual irradiance simulation
            with open(tstep_file) as tf:
                timestep = int(tf.readline())
            has_t_step = True

        # parse the sun-up-hours
        sun_up_hours = [int(h * timestep) for h in sun_up_hours]

        # create the header that will be used for all of the data collections
        aper = AnalysisPeriod(timestep=timestep)
        if 'direct_sun_hours' in res_folder:
            head = Header(Time(), 'hr', aper)
        elif has_t_step:
            head = Header(Irradiance(), 'W/m2', aper)
        else:
            head = Header(Illuminance(), 'lux', aper)
        dgp_head = Header(Fraction(), 'fraction', aper, metadata={'type': 'Daylight Glare Probability (DGP)'})

        # create the data collections from the .ill files
        data = []
        for grid_info, p_filt in zip(grids, pt_filter):
            grid_id = grid_info['full_id']
            ill_file = os.path.join(res_folder, '%s.ill' % grid_id)
            dgp_file = os.path.join(res_folder, '%s.dgp' % grid_id)
            if os.path.isfile(dgp_file):
                data_list = file_to_data(dgp_file, p_filt, sun_up_hours, dgp_head, timestep, grid_id)
            else:
                data_list = file_to_data(ill_file, p_filt, sun_up_hours, head, timestep, grid_id)
            data.append(data_list)
        data = list_to_data_tree(data)



# Honeybee: A Plugin for Environmental Analysis (GPL) started by Mostapha Sadeghipour Roudsari
# This file is part of Honeybee.
#
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a StateGeometry object that can be assigned to the shades_ of a dynamic
state using the "HB Dynamic State" component.
-

    Args:
        _geo: Rhino Brep or Mesh geometry to be converted to StateGeometry.
        _name_: Text to set the name for the StateGeometry and to be incorporated into
            unique StateGeometry identifier. If the name is not provided, a random name
            will be assigned.
        _modifier_: A Honeybee Radiance Modifier object for the geometry. If None,
            it will be the Generic Exterior Shade modifier in the lib. (Default: None).
    
    Returns:
        geo: A Honeybee StateGeometry object representing planar geometry that
            can be assigned to Radiance states. This can be assigned using the
            "HB Dynamic State" component.
"""

ghenv.Component.Name = 'HB Dynamic State Geometry'
ghenv.Component.NickName = 'StateGeo'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import uuid

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_rad_string, clean_rad_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_radiance.dynamic import StateGeometry
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.togeometry import to_face3d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    geo = []  # list of geometries that will be returned

    # set default name
    name = clean_and_id_rad_string('StateGeo') if _name_ is None \
        else clean_and_id_rad_string(_name_)

    # create the StateGeometry
    i = 0  # iterator to ensure each geometry gets a unique name
    for rh_geo in _geo:
        for lb_face in to_face3d(rh_geo):
            hb_geo = StateGeometry('{}_{}'.format(name, i), lb_face, _modifier_)
            if _name_ is not None:
                hb_geo.display_name = _name_
            geo.append(hb_geo)


# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply a section plane to a Honeybee Radiance View.
_
The plane will always be perpendicular to the view direction for perspective
and parallel view types. For fisheye view types, the clipping plane is actually
a clipping sphere, centered on the view point.
_
Objects in front of this imaginary plane will not be visible. This may be useful
for seeing through walls (to get a longer perspective from an exterior view point)
or for incremental rendering.
-

    Args:
        _view: A Honeybee Radiance View object to which a section plane should be applied.
        _origin: An point to set the origin of the section plane in 3D space. Note
            that the section plane is always perpenicular to the view direction
            for perspective and parallel views.

    Returns:
        view: A Honeybee View object that can be used in a view-based recipe.
"""

ghenv.Component.Name = 'HB Section Plane View'
ghenv.Component.NickName = 'SectionView'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_radiance.view import View
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # process the input view and origin
    if isinstance(_view, str):
        view = View.from_string(_view)
    else:
        assert isinstance(_view, View), \
            'Expected Honeybee View. Got {}.'.format(type(_view))
        view = _view.duplicate()
    view_pt = Point3D(*_view.position)
    origin = to_point3d(_origin)

    # set the fore clip according to the distance
    view.fore_clip = origin.distance_to_point(view_pt)
    #view.position = (origin.x, origin.y, origin.z)



# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate Spatial Daylight Autonomy (sDA) from lists of daylight autonomy values.
_
As per IES-LM-83-12 Spatial Daylight Autonomy (sDA) is a metric describing
annual sufficiency of ambient daylight levels in interior environments.
It is defined as the percent of an analysis area (the area where calcuations
are performed -typically across an entire space) that meets a minimum
daylight illuminance level for a specified fraction of the operating hours
per year. The sDA value is expressed as a percentage of area.
_
Note: This component will only output a LEED compliant sDA if you've run the
simulation with dynamic blinds and blind schedules as per the IES-LM-83-12
standard. If you are not using dynamic blinds, then this output is NOT LEED
compliant.

-
    Args:
        _DA: A data tree of daylight autonomy values output from the "HB Annual Dalyight"
            recipe or the "HB Annual Daylight Metrics" component. Note that,
            unless these DA values follow LM83 dynamic blinds setup, the resulting
            sDA is not LEED compliant.
        mesh_: An optional list of Meshes that align with the _DA data tree above, which
            will be used to assign an area to each sensor. If no mesh is connected
            here, it will be assumed that each sensor represents an equal area
            to all of the others.
        _target_time_: A minimum threshold of occupied time (eg. 50% of the time), above
            which a given sensor passes and contributes to the spatial daylight
            autonomy. (Default: 50%).

    Returns:
        report: Reports, errors, warnings, etc.
        sDA: Spatial daylight autonomy as percentage of area for each analysis grid.
        pass_fail: A data tree of zeros and ones, which indicate whether a given senor
            passes the criteria for being daylit (1) or fails the criteria (0). 
            Each value is for a different sensor of the grid. These can be plugged
            into the "LB Spatial Heatmap" component along with meshes of the
            sensor grids to visualize results.
"""

ghenv.Component.Name = 'HB Spatial Daylight Autonomy'
ghenv.Component.NickName = 'sDA'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug_rhino.togeometry import to_mesh3d
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        data_tree_to_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # process the input values into a rokable format
    da_mtx = [item[-1] for item in data_tree_to_list(_DA)]
    _target_time_ = 50 if _target_time_ is None else _target_time_
    lb_meshes = [to_mesh3d(mesh) for mesh in mesh_]

    # determine whether each point passes or fails
    pass_fail = [[int(val > _target_time_) for val in grid] for grid in da_mtx]

    # compute spatial daylight autonomy from the pass/fail results
    if len(lb_meshes) == 0:  # all sensors represent the same area
        sDA = [sum(pf_list) / len(pf_list) for pf_list in pass_fail]
    else:  # weight the sensors based on the area of mesh faces
        sDA = []
        for i, mesh in enumerate(lb_meshes):
            m_area = mesh.area
            weights = [fa / m_area for fa in mesh.face_areas]
            sDA.append(sum(v * w for v, w in zip(pass_fail[i], weights)))

    pass_fail = list_to_data_tree(pass_fail)  # convert matrix to data tree
